@@ -12,7 +12,6 @@ from albucore.decorators import contiguous, preserve_channel_dim
 from albucore.utils import (
     MAX_OPENCV_WORKING_CHANNELS,
     MAX_VALUES_BY_DTYPE,
-    MONO_CHANNEL_DIMENSIONS,
     NormalizationType,
     ValueType,
     clip,
@@ -32,10 +31,10 @@ def add_weighted_simsimd(img1: np.ndarray, weight1: float, img2: np.ndarray, wei
     original_dtype = img1.dtype
 
     if img2.dtype != original_dtype:
-        img2 = clip(img2.astype(original_dtype), original_dtype, inplace=True)
+        img2 = clip(img2.astype(original_dtype, copy=False), original_dtype, inplace=True)
 
     return np.frombuffer(
-        ss.wsum(img1.reshape(-1), img2.astype(original_dtype).reshape(-1), alpha=weight1, beta=weight2),
+        ss.wsum(img1.reshape(-1), img2.astype(original_dtype, copy=False).reshape(-1), alpha=weight1, beta=weight2),
         dtype=original_dtype,
     ).reshape(
         original_shape,
@@ -51,7 +50,7 @@ def multiply_by_constant_simsimd(img: np.ndarray, value: float) -> np.ndarray:
 
 
 def add_constant_simsimd(img: np.ndarray, value: float) -> np.ndarray:
-    return add_weighted_simsimd(img, 1, (np.ones_like(img) * value).astype(img.dtype), 1)
+    return add_weighted_simsimd(img, 1, (np.ones_like(img) * value).astype(img.dtype, copy=False), 1)
 
 
 def create_lut_array(
@@ -92,11 +91,12 @@ def apply_lut(
 
     if isinstance(value, (int, float)):
         lut = create_lut_array(dtype, value, operation)
-        return sz_lut(img, clip(lut, dtype), inplace)
+        return sz_lut(img, clip(lut, dtype, inplace=False), False)
 
     num_channels = img.shape[-1]
-    luts = create_lut_array(dtype, value, operation)
-    return cv2.merge([sz_lut(img[:, :, i], clip(luts[i], dtype, inplace=False), inplace) for i in range(num_channels)])
+
+    luts = clip(create_lut_array(dtype, value, operation), dtype, inplace=False)
+    return cv2.merge([sz_lut(img[:, :, i], luts[i], inplace) for i in range(num_channels)])
 
 
 def prepare_value_opencv(
@@ -135,14 +135,14 @@ def _prepare_array_value(
     operation: Literal["add", "multiply"],
 ) -> np.ndarray:
     if value.dtype == np.float64:
-        value = value.astype(np.float32)
+        value = value.astype(np.float32, copy=False)
     if value.ndim == 1:
         value = value.reshape(1, 1, -1)
     value = np.broadcast_to(value, img.shape)
     if operation == "add" and img.dtype == np.uint8:
         if np.all(value >= 0):
-            return clip(value, np.uint8)
-        return np.trunc(value).astype(np.float32)
+            return clip(value, np.uint8, inplace=False)
+        return np.trunc(value).astype(np.float32, copy=False)
     return value
 
 
@@ -154,7 +154,7 @@ def apply_numpy(
     if operation == "add" and img.dtype == np.uint8:
         value = np.int16(value)
 
-    return np_operations[operation](img.astype(np.float32), value)
+    return np_operations[operation](img.astype(np.float32, copy=False), value)
 
 
 def multiply_lut(img: np.ndarray, value: np.ndarray | float, inplace: bool) -> np.ndarray:
@@ -165,7 +165,7 @@ def multiply_lut(img: np.ndarray, value: np.ndarray | float, inplace: bool) -> n
 def multiply_opencv(img: np.ndarray, value: np.ndarray | float) -> np.ndarray:
     value = prepare_value_opencv(img, value, "multiply")
     if img.dtype == np.uint8:
-        return cv2.multiply(img.astype(np.float32), value)
+        return cv2.multiply(img.astype(np.float32, copy=False), value)
     return cv2.multiply(img, value)
 
 
@@ -223,7 +223,10 @@ def add_opencv(img: np.ndarray, value: np.ndarray | float, inplace: bool = False
     )
 
     if needs_float:
-        return cv2.add(img.astype(np.float32), value if isinstance(value, (int, float)) else value.astype(np.float32))
+        return cv2.add(
+            img.astype(np.float32, copy=False),
+            value if isinstance(value, (int, float)) else value.astype(np.float32, copy=False),
+        )
 
     # Use img as the destination array if inplace=True
     dst = img if inplace else None
@@ -272,25 +275,24 @@ def add(img: np.ndarray, value: ValueType, inplace: bool = False) -> np.ndarray:
 
 
 def normalize_numpy(img: np.ndarray, mean: float | np.ndarray, denominator: float | np.ndarray) -> np.ndarray:
-    img = img.astype(np.float32)
+    img = img.astype(np.float32, copy=False)
+    # Ensure mean and denominator are float32 to avoid dtype promotion
+    mean = mean.astype(np.float32, copy=False) if isinstance(mean, np.ndarray) else np.float32(mean)
+    denominator = (
+        denominator.astype(np.float32, copy=False) if isinstance(denominator, np.ndarray) else np.float32(denominator)
+    )
     img -= mean
-    return img * denominator
+    return (img * denominator).astype(np.float32, copy=False)
 
 
 @preserve_channel_dim
 def normalize_opencv(img: np.ndarray, mean: float | np.ndarray, denominator: float | np.ndarray) -> np.ndarray:
-    img = img.astype(np.float32)
+    img = img.astype(np.float32, copy=False)
     mean_img = np.zeros_like(img, dtype=np.float32)
     denominator_img = np.zeros_like(img, dtype=np.float32)
 
-    # If mean or denominator are scalar, convert them to arrays
-    if isinstance(mean, (float, int)):
-        mean = np.full(img.shape, mean, dtype=np.float32)
-    if isinstance(denominator, (float, int)):
-        denominator = np.full(img.shape, denominator, dtype=np.float32)
-
     # Ensure the shapes match for broadcasting
-    mean_img = (mean_img + mean).astype(np.float32)
+    mean_img = (mean_img + mean).astype(np.float32, copy=False)
     denominator_img = denominator_img + denominator
 
     result = cv2.subtract(img, mean_img)
@@ -304,27 +306,39 @@ def normalize_lut(img: np.ndarray, mean: float | np.ndarray, denominator: float
     num_channels = get_num_channels(img)
 
     if isinstance(denominator, (float, int)) and isinstance(mean, (float, int)):
-        lut = (np.arange(0, max_value + 1, dtype=np.float32) - mean) * denominator
+        lut = ((np.arange(0, max_value + 1, dtype=np.float32) - mean) * denominator).astype(np.float32)
         return cv2.LUT(img, lut)
 
-    if isinstance(denominator, np.ndarray) and denominator.shape != ():
-        denominator = denominator.reshape(-1, 1)
-
+    # Convert to float32 if needed
     if isinstance(mean, np.ndarray):
-        mean = mean.reshape(-1, 1)
+        mean = mean.astype(np.float32, copy=False)
+    if isinstance(denominator, np.ndarray):
+        denominator = denominator.astype(np.float32, copy=False)
 
-    luts = (np.arange(0, max_value + 1, dtype=np.float32) - mean) * denominator
+    # Vectorized LUT creation - shape: (256, num_channels)
+    arange_vals = np.arange(0, max_value + 1, dtype=np.float32)
+    luts = (arange_vals[:, np.newaxis] - mean) * denominator
 
-    return cv2.merge([cv2.LUT(img[:, :, i], luts[i]) for i in range(num_channels)])
+    # Pre-allocate result array
+    result = np.empty_like(img, dtype=np.float32)
+    for i in range(num_channels):
+        result[..., i] = cv2.LUT(img[..., i], luts[:, i])
+
+    return result
 
 
 def normalize(img: np.ndarray, mean: ValueType, denominator: ValueType) -> np.ndarray:
     num_channels = get_num_channels(img)
     denominator = convert_value(denominator, num_channels)
     mean = convert_value(mean, num_channels)
+
     if img.dtype == np.uint8:
         return normalize_lut(img, mean, denominator)
 
+    if img.dtype == np.float32:
+        return normalize_numpy(img, mean, denominator)
+
+    # Fallback to OpenCV for other dtypes
     return normalize_opencv(img, mean, denominator)
 
 
@@ -343,7 +357,7 @@ def power_opencv(img: np.ndarray, value: float) -> np.ndarray:
         return cv2.pow(img, value)
     if img.dtype == np.uint8 and isinstance(value, float):
         # For uint8 images, convert to float32, apply power, then convert back to uint8
-        img_float = img.astype(np.float32)
+        img_float = img.astype(np.float32, copy=False)
         return cv2.pow(img_float, value)
 
     raise ValueError(f"Unsupported image type {img.dtype} for power operation with value {value}")
@@ -368,7 +382,7 @@ def power(img: np.ndarray, exponent: ValueType, inplace: bool = False) -> np.nda
 
 
 def add_weighted_numpy(img1: np.ndarray, weight1: float, img2: np.ndarray, weight2: float) -> np.ndarray:
-    return img1.astype(np.float32) * weight1 + img2.astype(np.float32) * weight2
+    return img1.astype(np.float32, copy=False) * weight1 + img2.astype(np.float32, copy=False) * weight2
 
 
 @preserve_channel_dim
@@ -430,7 +444,7 @@ def multiply_add_opencv(img: np.ndarray, factor: ValueType, value: ValueType) ->
     if isinstance(value, (int, float)) and value == 0 and isinstance(factor, (int, float)) and factor == 0:
         return np.zeros_like(img)
 
-    result = img.astype(np.float32)
+    result = img.astype(np.float32, copy=False)
     result = (
         cv2.multiply(result, np.ones_like(result) * factor, dtype=cv2.CV_64F)
         if factor != 0
@@ -471,140 +485,323 @@ def multiply_add(img: np.ndarray, factor: ValueType, value: ValueType, inplace:
     return multiply_add_opencv(img, factor, value)
 
 
-@preserve_channel_dim
-def normalize_per_image_opencv(img: np.ndarray, normalization: NormalizationType) -> np.ndarray:
-    img = img.astype(np.float32)
+def _compute_image_stats_opencv(img: np.ndarray) -> tuple[float, float]:
+    """Compute global mean and std for an image."""
     eps = 1e-4
+    if img.ndim > 3:
+        # For 4D/5D arrays (video/volume), OpenCV returns global mean/std directly
+        mean, std = cv2.meanStdDev(img)
+        return float(mean[0, 0]), float(std[0, 0]) + eps
+    # For 3D images, use numpy for accurate global statistics
+    return float(img.mean()), float(img.std()) + eps
 
-    if img.ndim == MONO_CHANNEL_DIMENSIONS:
-        img = np.expand_dims(img, axis=-1)
 
-    if normalization == "image" or (img.shape[-1] == 1 and normalization == "image_per_channel"):
-        mean = img.mean().item()
-        std = img.std().item() + eps
-        if img.shape[-1] > MAX_OPENCV_WORKING_CHANNELS:
-            mean = np.full_like(img, mean)
-            std = np.full_like(img, std)
-        normalized_img = cv2.divide(cv2.subtract(img, mean), std)
-        return normalized_img.clip(-20, 20)
-
-    if normalization == "image_per_channel":
+def _compute_per_channel_stats_opencv(img: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
+    """Compute per-channel mean and std."""
+    eps = 1e-4
+    if img.ndim > 3:
+        # For 4D/5D arrays, compute per-channel statistics using numpy
+        axes = tuple(range(img.ndim - 1))  # All axes except channel
+        mean = img.mean(axis=axes)
+        std = img.std(axis=axes) + eps
+    else:
+        # For 3D arrays, use OpenCV
         mean, std = cv2.meanStdDev(img)
         mean = mean[:, 0]
-        std = std[:, 0]
+        std = std[:, 0] + eps
+    return mean, std
 
-        if img.shape[-1] > MAX_OPENCV_WORKING_CHANNELS:
-            mean = np.full_like(img, mean)
-            std = np.full_like(img, std)
 
-        normalized_img = cv2.divide(cv2.subtract(img, mean), std, dtype=cv2.CV_32F)
-        return normalized_img.clip(-20, 20)
+def _normalize_mean_std_opencv(img_f: np.ndarray, mean: float | np.ndarray, std: float | np.ndarray) -> np.ndarray:
+    """Apply mean-std normalization using OpenCV or NumPy based on dimensionality."""
+    if img_f.ndim > 3:
+        # Use NumPy operations for 4D/5D (faster)
+        normalized_img = (img_f - mean) / std
+    else:
+        # Use OpenCV for 3D
+        if img_f.shape[-1] > MAX_OPENCV_WORKING_CHANNELS:
+            mean = np.full_like(img_f, mean)
+            std = np.full_like(img_f, std)
+        normalized_img = cv2.divide(cv2.subtract(img_f, mean, dtype=cv2.CV_32F), std, dtype=cv2.CV_32F)
+    return np.clip(normalized_img, -20, 20, out=normalized_img)
 
-    if normalization == "min_max" or (img.shape[-1] == 1 and normalization == "min_max_per_channel"):
-        img_min = img.min()
-        img_max = img.max()
+
+def _normalize_min_max_per_channel_opencv(img: np.ndarray) -> np.ndarray:
+    """Apply per-channel min-max normalization."""
+    eps = 1e-4
+    axes = tuple(range(img.ndim - 1))  # All axes except channel
+
+    img_min = img.min(axis=axes)
+    img_max = img.max(axis=axes)
+
+    if img.shape[-1] > MAX_OPENCV_WORKING_CHANNELS:
+        img_min = np.full_like(img, img_min)
+        img_max = np.full_like(img, img_max)
+
+    # Use NumPy operations for 4D/5D (faster), OpenCV for 3D
+    if img.ndim > 3:
+        normalized_img = (img - img_min) / (img_max - img_min + eps)
+    else:
+        normalized_img = cv2.divide(cv2.subtract(img, img_min), (img_max - img_min + eps), dtype=cv2.CV_32F)
+
+    return np.clip(normalized_img, -20, 20, out=normalized_img)
+
+
+@preserve_channel_dim
+def normalize_per_image_opencv(
+    img: np.ndarray,
+    normalization: NormalizationType,
+) -> np.ndarray:
+    """Normalize an image using OpenCV operations based on the specified normalization type.
+
+    This function normalizes an image using various strategies, optimized with OpenCV operations
+    for better performance on standard image types.
+
+    Args:
+        img: Input image as a numpy array with shape (H, W, C).
+        normalization: Type of normalization to apply. Options are:
+            - "image": Normalize using global mean and std across all pixels
+            - "image_per_channel": Normalize each channel separately using its own mean and std
+            - "min_max": Scale to [0, 1] using global min and max values
+            - "min_max_per_channel": Scale each channel to [0, 1] using per-channel min and max
+
+    Returns:
+        Normalized image as float32 array with values clipped to [-20, 20] range to prevent
+        extreme values that could cause training instability.
+
+    Raises:
+        ValueError: If an unknown normalization method is specified.
+
+    Notes:
+        - The function automatically converts input to float32
+        - Adds epsilon (1e-4) to std deviation to prevent division by zero
+        - For images with >4 channels, falls back to array operations as OpenCV has limitations
+        - Single channel images treated as "image" normalization when "image_per_channel" is specified
+    """
+    # Handle single-channel edge case
+    if img.shape[-1] == 1 and normalization == "image_per_channel":
+        normalization = "image"
+    if img.shape[-1] == 1 and normalization == "min_max_per_channel":
+        normalization = "min_max"
+
+    if normalization == "image":
+        mean, std = _compute_image_stats_opencv(img)
+        img_f = img.astype(np.float32, copy=False)
+        return _normalize_mean_std_opencv(img_f, mean, std)
+
+    if normalization == "image_per_channel":
+        mean, std = _compute_per_channel_stats_opencv(img)
+        img_f = img.astype(np.float32, copy=False)
+        return _normalize_mean_std_opencv(img_f, mean, std)
+
+    if normalization == "min_max":
         return cv2.normalize(img, None, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F)
 
     if normalization == "min_max_per_channel":
-        img_min = img.min(axis=(0, 1))
-        img_max = img.max(axis=(0, 1))
-
-        if img.shape[-1] > MAX_OPENCV_WORKING_CHANNELS:
-            img_min = np.full_like(img, img_min)
-            img_max = np.full_like(img, img_max)
-
-        return cv2.divide(cv2.subtract(img, img_min), (img_max - img_min + eps), dtype=cv2.CV_32F).clip(-20, 20)
+        return _normalize_min_max_per_channel_opencv(img)
 
     raise ValueError(f"Unknown normalization method: {normalization}")
 
 
 @preserve_channel_dim
-def normalize_per_image_numpy(img: np.ndarray, normalization: NormalizationType) -> np.ndarray:
-    img = img.astype(np.float32)
-    eps = 1e-4
+def normalize_per_image_numpy(
+    img: np.ndarray,
+    normalization: NormalizationType,
+) -> np.ndarray:
+    """Normalize an image using pure NumPy operations based on the specified normalization type.
 
-    if img.ndim == MONO_CHANNEL_DIMENSIONS:
-        img = np.expand_dims(img, axis=-1)
+    This function provides the same normalization strategies as normalize_per_image_opencv but uses
+    pure NumPy operations. This can be useful for compatibility or when OpenCV is not available.
+
+    Args:
+        img: Input image as a numpy array with shape (H, W, C).
+        normalization: Type of normalization to apply. Options are:
+            - "image": Normalize using global mean and std across all pixels
+            - "image_per_channel": Normalize each channel separately using its own mean and std
+            - "min_max": Scale to [0, 1] using global min and max values
+            - "min_max_per_channel": Scale each channel to [0, 1] using per-channel min and max
+
+    Returns:
+        Normalized image as float32 array with values clipped to [-20, 20] range to prevent
+        extreme values that could cause training instability.
+
+    Raises:
+        ValueError: If an unknown normalization method is specified.
+
+    Notes:
+        - The function automatically converts input to float32
+        - Adds epsilon (1e-4) to std deviation to prevent division by zero
+        - Uses in-place operations where possible for memory efficiency
+        - Generally slower than the OpenCV version but more portable
+    """
+    img = img.astype(np.float32, copy=False)
+    eps = 1e-4
 
     if normalization == "image":
         mean = img.mean()
         std = img.std() + eps
         normalized_img = (img - mean) / std
-        return normalized_img.clip(-20, 20)
+        return np.clip(normalized_img, -20, 20, out=normalized_img)
 
     if normalization == "image_per_channel":
-        pixel_mean = img.mean(axis=(0, 1))
-        pixel_std = img.std(axis=(0, 1)) + eps
+        axes = tuple(range(img.ndim - 1))  # All axes except channel
+        pixel_mean = img.mean(axis=axes)
+        pixel_std = img.std(axis=axes) + eps
         normalized_img = (img - pixel_mean) / pixel_std
-        return normalized_img.clip(-20, 20)
+        return np.clip(normalized_img, -20, 20, out=normalized_img)
 
     if normalization == "min_max":
         img_min = img.min()
         img_max = img.max()
-        return (img - img_min) / (img_max - img_min + eps)
+        return np.clip((img - img_min) / (img_max - img_min + eps), -20, 20, out=img)
 
     if normalization == "min_max_per_channel":
-        img_min = img.min(axis=(0, 1))
-        img_max = img.max(axis=(0, 1))
-        return (img - img_min) / (img_max - img_min + eps)
+        axes = tuple(range(img.ndim - 1))  # All axes except channel
+        img_min = img.min(axis=axes)
+        img_max = img.max(axis=axes)
+        return np.clip((img - img_min) / (img_max - img_min + eps), -20, 20, out=img)
 
     raise ValueError(f"Unknown normalization method: {normalization}")
 
 
 @preserve_channel_dim
-def normalize_per_image_lut(img: np.ndarray, normalization: NormalizationType) -> np.ndarray:
+def normalize_per_image_lut(
+    img: np.ndarray,
+    normalization: NormalizationType,
+) -> np.ndarray:
+    """Normalize an image using lookup tables (LUT) for optimized performance on uint8 images.
+
+    This function implements the same normalization strategies but uses pre-computed lookup tables
+    for extremely fast normalization of uint8 images. This is the fastest method for uint8 data.
+
+    Args:
+        img: Input image as a numpy array with uint8 dtype and shape (H, W, C).
+        normalization: Type of normalization to apply. Options are:
+            - "image": Normalize using global mean and std across all pixels
+            - "image_per_channel": Normalize each channel separately using its own mean and std
+            - "min_max": Scale to [0, 1] using global min and max values
+            - "min_max_per_channel": Scale each channel to [0, 1] using per-channel min and max
+
+    Returns:
+        Normalized image as float32 array with values clipped to [-20, 20] range to prevent
+        extreme values that could cause training instability.
+
+    Raises:
+        ValueError: If an unknown normalization method is specified.
+
+    Notes:
+        - Designed specifically for uint8 images for maximum performance
+        - Creates a 256-element lookup table mapping each possible uint8 value to its normalized value
+        - Uses OpenCV's LUT function for fast application of the transformation
+        - For per-channel normalization, creates separate LUTs for each channel
+        - Single channel images treated as "image" normalization when "image_per_channel" is specified
+    """
     dtype = img.dtype
     max_value = MAX_VALUES_BY_DTYPE[dtype]
     eps = 1e-4
     num_channels = get_num_channels(img)
 
-    if img.ndim == MONO_CHANNEL_DIMENSIONS:
-        img = np.expand_dims(img, axis=-1)
-
     if normalization == "image" or (img.shape[-1] == 1 and normalization == "image_per_channel"):
-        mean = img.mean()
-        std = img.std() + eps
-        lut = (np.arange(0, max_value + 1, dtype=np.float32) - mean) / std
+        if img.ndim > 3:
+            # For 4D/5D arrays (video/volume), OpenCV returns global mean/std directly
+            mean, std = cv2.meanStdDev(img)
+            mean = mean[0, 0]
+            std = std[0, 0] + eps
+        else:
+            # For 3D images, use numpy for accurate global statistics
+            mean = img.mean()
+            std = img.std() + eps
+
+        lut = ((np.arange(0, max_value + 1, dtype=np.float32) - mean) / std).astype(np.float32)
         return cv2.LUT(img, lut).clip(-20, 20)
 
     if normalization == "image_per_channel":
-        pixel_mean = img.mean(axis=(0, 1))
-        pixel_std = img.std(axis=(0, 1)) + eps
-        luts = [
-            (np.arange(0, max_value + 1, dtype=np.float32) - pixel_mean[c]) / pixel_std[c] for c in range(num_channels)
-        ]
-        return cv2.merge([cv2.LUT(img[:, :, i], luts[i]).clip(-20, 20) for i in range(num_channels)])
+        axes = tuple(range(img.ndim - 1))  # All axes except channel
+        pixel_mean = img.mean(axis=axes).astype(np.float32)
+        pixel_std = img.std(axis=axes).astype(np.float32) + np.float32(eps)
+
+        # Create all LUTs at once using vectorized operations
+        arange_vals = np.arange(0, max_value + 1, dtype=np.float32)
+        # LUTs shape will be (256, num_channels)
+        luts = (arange_vals[:, np.newaxis] - pixel_mean) / pixel_std
+
+        result = np.empty_like(img, dtype=np.float32)
+        for i in range(num_channels):
+            result[..., i] = cv2.LUT(img[..., i], luts[:, i])
+        return result.clip(-20, 20)
 
     if normalization == "min_max" or (img.shape[-1] == 1 and normalization == "min_max_per_channel"):
         img_min = img.min()
         img_max = img.max()
-        lut = (np.arange(0, max_value + 1, dtype=np.float32) - img_min) / (img_max - img_min + eps)
-        return cv2.LUT(img, lut)
+        lut = ((np.arange(0, max_value + 1, dtype=np.float32) - img_min) / (img_max - img_min + eps)).astype(np.float32)
+        return cv2.LUT(img, lut).clip(-20, 20)
 
     if normalization == "min_max_per_channel":
-        img_min = img.min(axis=(0, 1))
-        img_max = img.max(axis=(0, 1))
-        luts = [
-            (np.arange(0, max_value + 1, dtype=np.float32) - img_min[c]) / (img_max[c] - img_min[c] + eps)
-            for c in range(num_channels)
-        ]
+        axes = tuple(range(img.ndim - 1))  # All axes except channel
+        img_min = img.min(axis=axes)
+        img_max = img.max(axis=axes)
 
-        return cv2.merge([cv2.LUT(img[:, :, i], luts[i]) for i in range(num_channels)])
+        # Create all LUTs at once using vectorized operations
+        arange_vals = np.arange(0, max_value + 1, dtype=np.float32)
+        # LUTs shape will be (256, num_channels)
+        luts = ((arange_vals[:, np.newaxis] - img_min) / (img_max - img_min + eps)).astype(np.float32)
+
+        result = np.empty_like(img, dtype=np.float32)
+        for i in range(num_channels):
+            result[..., i] = cv2.LUT(img[..., i], luts[:, i])
+        return result.clip(-20, 20)
 
     raise ValueError(f"Unknown normalization method: {normalization}")
 
 
 def normalize_per_image(img: np.ndarray, normalization: NormalizationType) -> np.ndarray:
-    if img.dtype == np.uint8 and normalization != "per_image_per_channel":
+    """Normalize an image using the most efficient method based on image dtype and normalization type.
+
+    This is the main entry point for image normalization that automatically selects the optimal
+    implementation based on the input image data type.
+
+    Args:
+        img: Input image as a numpy array with shape (H, W, C), (N, H, W, C), or (N, D, H, W, C).
+        normalization: Type of normalization to apply. Options are:
+            - "image": Normalize using global mean and std across all pixels
+            - "image_per_channel": Normalize each channel separately using its own mean and std
+            - "min_max": Scale to [0, 1] using global min and max values
+            - "min_max_per_channel": Scale each channel to [0, 1] using per-channel min and max
+
+    Returns:
+        Normalized image as float32 array with values clipped to [-20, 20] range.
+
+    Notes:
+        - For uint8 images (except "min_max"), uses LUT method for maximum speed
+        - For other dtypes, uses OpenCV or NumPy implementation for good performance
+        - Automatically determines spatial axes based on input dimensions
+    """
+    # Route uint8 images
+    if img.dtype == np.uint8:
+        # Use LUT for everything except min_max (where OpenCV is 3x faster)
+        if normalization == "min_max":
+            return normalize_per_image_opencv(img, normalization)
+        # LUT is fastest for "image", "image_per_channel", and "min_max_per_channel"
         return normalize_per_image_lut(img, normalization)
 
+    # Route float32 images
+    if img.dtype == np.float32:
+        if normalization == "image":
+            # NumPy is 1.5x faster for "image" normalization
+            return normalize_per_image_numpy(img, normalization)
+        # OpenCV is fastest or equal for all other normalizations
+        return normalize_per_image_opencv(img, normalization)
+
+    # Default fallback: OpenCV for single images, NumPy for videos/volumes
+    if img.ndim > 3:
+        return normalize_per_image_numpy(img, normalization)
     return normalize_per_image_opencv(img, normalization)
 
 
 def to_float_numpy(img: np.ndarray, max_value: float | None = None) -> np.ndarray:
     if max_value is None:
         max_value = get_max_value(img.dtype)
-    return (img / max_value).astype(np.float32)
+    return (img / max_value).astype(np.float32, copy=False)
 
 
 @preserve_channel_dim
@@ -612,7 +809,7 @@ def to_float_opencv(img: np.ndarray, max_value: float | None = None) -> np.ndarr
     if max_value is None:
         max_value = get_max_value(img.dtype)
 
-    img_float = img.astype(np.float32)
+    img_float = img.astype(np.float32, copy=False)
 
     num_channels = get_num_channels(img)
 
@@ -638,7 +835,7 @@ def to_float_lut(img: np.ndarray, max_value: float | None = None) -> np.ndarray:
 
 def to_float(img: np.ndarray, max_value: float | None = None) -> np.ndarray:
     if img.dtype == np.float64:
-        return img.astype(np.float32)
+        return img.astype(np.float32, copy=False)
     if img.dtype == np.float32:
         return img
     if img.dtype == np.uint8:
@@ -657,17 +854,17 @@ def from_float_opencv(img: np.ndarray, target_dtype: np.dtype, max_value: float
     if max_value is None:
         max_value = get_max_value(target_dtype)
 
-    img_float = img.astype(np.float32)
+    img_float = img.astype(np.float32, copy=False)
 
     num_channels = get_num_channels(img)
 
     if num_channels > MAX_OPENCV_WORKING_CHANNELS:
         # For images with more than 4 channels, create a full-sized multiplier
         max_value_array = np.full_like(img_float, max_value)
-        return clip(np.rint(cv2.multiply(img_float, max_value_array)), target_dtype)
+        return clip(np.rint(cv2.multiply(img_float, max_value_array)), target_dtype, inplace=False)
 
     # For images with 4 or fewer channels, use scalar multiplication
-    return clip(np.rint(img * max_value), target_dtype)
+    return clip(np.rint(img * max_value), target_dtype, inplace=False)
 
 
 def from_float(img: np.ndarray, target_dtype: np.dtype, max_value: float | None = None) -> np.ndarray:
@@ -695,7 +892,7 @@ def from_float(img: np.ndarray, target_dtype: np.dtype, max_value: float | None
         return img
 
     if target_dtype == np.float64:
-        return img.astype(np.float32)
+        return img.astype(np.float32, copy=False)
 
     if img.dtype == np.float32:
         return from_float_opencv(img, target_dtype, max_value)
@@ -710,6 +907,9 @@ def hflip_numpy(img: np.ndarray) -> np.ndarray:
 
 @preserve_channel_dim
 def hflip_cv2(img: np.ndarray) -> np.ndarray:
+    # OpenCV's flip function has a limitation of 512 channels
+    if img.ndim > 2 and img.shape[2] > 512:
+        return _flip_multichannel(img, flip_code=1)
     return cv2.flip(img, 1)
 
 
@@ -719,6 +919,9 @@ def hflip(img: np.ndarray) -> np.ndarray:
 
 @preserve_channel_dim
 def vflip_cv2(img: np.ndarray) -> np.ndarray:
+    # OpenCV's flip function has a limitation of 512 channels
+    if img.ndim > 2 and img.shape[2] > 512:
+        return _flip_multichannel(img, flip_code=0)
     return cv2.flip(img, 0)
 
 
@@ -731,6 +934,48 @@ def vflip(img: np.ndarray) -> np.ndarray:
     return vflip_cv2(img)
 
 
+def _flip_multichannel(img: np.ndarray, flip_code: int) -> np.ndarray:
+    """Process images with more than 512 channels by splitting into chunks.
+
+    OpenCV's flip function has a limitation where it can only handle images with up to 512 channels.
+    This function works around that limitation by splitting the image into chunks of 512 channels,
+    flipping each chunk separately, and then concatenating the results.
+
+    Args:
+        img: Input image with many channels
+        flip_code: OpenCV flip code (0 for vertical, 1 for horizontal, -1 for both)
+
+    Returns:
+        Flipped image with all channels preserved
+    """
+    # Get image dimensions
+    height, width = img.shape[:2]
+    num_channels = img.shape[2]
+
+    # If the image has fewer than 512 channels, use cv2.flip directly
+    if num_channels <= 512:
+        return cv2.flip(img, flip_code)
+
+    # Process in chunks of 512 channels
+    chunk_size = 512
+    result_chunks = []
+
+    for i in range(0, num_channels, chunk_size):
+        end_idx = min(i + chunk_size, num_channels)
+        chunk = img[:, :, i:end_idx]
+        flipped_chunk = cv2.flip(chunk, flip_code)
+
+        # Ensure the chunk maintains its dimensionality
+        # This is needed when the last chunk has only one channel and cv2.flip reduces the dimensions
+        if flipped_chunk.ndim == 2 and img.ndim == 3:
+            flipped_chunk = np.expand_dims(flipped_chunk, axis=2)
+
+        result_chunks.append(flipped_chunk)
+
+    # Concatenate the chunks along the channel dimension
+    return np.concatenate(result_chunks, axis=2)
+
+
 def float32_io(func: Callable[..., np.ndarray]) -> Callable[..., np.ndarray]:
     """Decorator to ensure float32 input/output for image processing functions.
 
